
""" The :mod:`dubious.discord` module contains the dataclasses for the Discord API's objects.

    * :mod:`.disc` is made to deal with transforming Discord objects to and from json.
    * :mod:`.api` defines the structures of all of the objects that Discord will send in Interactions.
    * :mod:`.req` defines the structures of all of the objects that Discord accepts/sends back from http requests to its API.
    
    """

from dubious.discord import disc, api, req