
""" DubiousDiscord is an opinionated Discord API wrapper written in Python.

    It features:

    * a full set of dataclasses automatically generated by parsing the `Discord API Documentation <https://github.com/discord/discord-api-docs>`_,
    * a decorator-based `Interaction <https://discord.com/developers/docs/interactions/application-commands>`_ builder through :class:`.pory.Pory`, and
    * a small DSL through :mod:`~dubious.zap` to make transforming information into and out of human-readable text (mostly Discord messages) easier.
    """

from dubious import pory, callback as cb, zap